/**
 * @file tuya_ai_toy.c
 * @brief
 * @version 0.1
 * @date 2024-11-01
 *
 * @copyright Copyright (c) 2024 Tuya Inc. All Rights Reserved.
 *
 * Permission is hereby granted, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), Under the premise of complying
 * with the license of the third-party open source software contained in the software,
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software.
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 */

#include <stdio.h>
#include "tuya_ai_toy.h"
#include "tkl_system.h"
#include "tuya_device_cfg.h"
#include "tal_log.h"
#include "tal_memory.h"
#include "tal_thread.h"
#include "tkl_audio.h"
#include "tal_queue.h"
#include "tal_system.h"
#include "tal_network.h"
#include "tal_mutex.h"
#include "tal_sw_timer.h"
#include "tkl_thread.h"
#include "tkl_gpio.h"
#include "base_event.h"
#include "tuya_wifi_status.h"
#include "ns/audio_subsys_ns.h"
#include "vad/audio_subsys_vad.h"
#include "tuya_ai_proc.h"
#include "tuya_ai_mp3_player.h"
#include "tuya_iot_wifi_api.h"
#include "tuya_ws_db.h"
#include "ty_cJSON.h"
#include "tuya_ai_battery.h"

#define AUDIO_SAMPLE_RATE          16000
#define SPK_SAMPLE_RATE            16000
#define AUDIO_SAMPLE_BITS          16
#define AUDIO_CHANNEL              1
#define AUDIO_20MS_LEN             (AUDIO_SAMPLE_RATE * 2 / 50)
#define AUDIO_100MS_LEN            (AUDIO_SAMPLE_RATE * 2 / 10)
#define AUDIO_ONE_SEC_LEN          (AUDIO_SAMPLE_RATE * 2)
#define AUDIO_TEN_SEC_LEN          (AUDIO_SAMPLE_RATE * 2 * 10)
#define AUDIO_BUFF_MAX_LEN         (AUDIO_ONE_SEC_LEN * 10)
#define AUDIO_TTS_STREAM_BUFF_MAX_LEN    (1024 * 64)
#define AUDIO_PCM_SLICE_BUFF_LEN   (640)
#define AUDIO_PCM_SLICE_TIME       (AUDIO_PCM_SLICE_BUFF_LEN / 2 / (AUDIO_SAMPLE_RATE / 1000))
#define SILENCE_THRESHOLD_ONE_SHORT_MODE    200     // 静音等待阈值，单位毫秒
#define ACTIVE_THRESHOLD_ONE_SHORT_MODE     200     // 语音活动触发阈值，单位毫秒
#define WAIT_STOP_PLAY_THRESHOLD            200     // 等待播放结束阈值，单位毫秒
#define SILENCE_THRESHOLD_HOLD_MODE         200     // 静音等待阈值，单位毫秒
#define ACTIVE_THRESHOLD_HOLD_MODE          200     // 语音活动触发阈值，单位毫秒
#define SPEECH_INACTIVE_HOLD_TIMEOUT        (100)   // 按键超时
#define SPEECH_ACTIVE_TIMEOUT               (10 * 1000)
#define AI_TOY_PARA "ai_toy_para"

typedef struct {
    TY_AI_TOY_PARA_T para;      // ai toy para
    BOOL_T is_init;             // ai toy init flag
    UINT_T silence_threshold;   // 静音等待阈值，单位毫秒
    UINT_T active_threshold;    // 语音活动触发阈值，单位毫秒
    UINT_T wait_stop_play_threshold; // 等待播放结束阈值，单位毫秒
    TY_AI_VoiceState voice_state;
    INT_T frame_duration_ms;    // 音频帧时长，单位毫秒
    GW_WIFI_NW_STAT_E nw_stat;  // 网络状态
    BOOL_T alert_flag;          // 提示标志
    UINT8_T power_state;        // 电源状态, 0~100
    UINT8_T volume;             // 音量, 0~100
    // add internal config here
    // ...
} TY_AI_TOY_CONFIG_T;

STATIC TY_AI_TOY_CONFIG_T s_ai_cfg;
STATIC THREAD_HANDLE s_ty_ai_toy_hander = NULL;
STATIC TY_AI_PROC_HANDLE s_ai_proc_handle = NULL;
STATIC INT_T s_vad_active = 0;
STATIC BOOL_T s_force_active = FALSE;
STATIC TIMER_ID s_speech_active_timer = NULL;
STATIC audio_subsys_module_t *vad_module = &audio_subsys_vad;
STATIC audio_subsys_module_t *ns_module = &audio_subsys_ns;

STATIC OPERATE_RET _report_sysinfo(VOID);

STATIC VOID _voice_stat_update(TY_AI_VoiceState stat)
{
    if (s_ai_cfg.voice_state == stat) {
        return;
    }
    s_ai_cfg.voice_state = stat;
    PR_ERR("move voice state to %d", s_ai_cfg.voice_state);
    ty_ai_voice_stat_post(s_ai_proc_handle, stat);
}

STATIC VOID_T _speech_active_timer_timeout(TIMER_ID timer_id, VOID_T *arg)
{
    s_force_active = FALSE;
    PR_ERR("speech active timer timeout, s_force_active = %d", s_force_active);
}

INT_T _event_play_status_cb(VOID_T *data)
{
    TUYA_AI_PLAYER_STAT stat = *(TUYA_AI_PLAYER_STAT *)data;

    PR_ERR("#### get stat-%d", stat);
    if (stat != TUYA_AI_PLAYER_STAT_STOP) {
        return 0;
    }

    // 监测player stop状态，重新触发下一轮对话录音
    if (s_ai_cfg.para.trigger_mode == TY_AI_TRIGGER_MODE_HOLD) {
        // 按键触发模式，播放结束后，需要重新触发下一轮对话录音
        PR_ERR("alert_flag = %d", s_ai_cfg.alert_flag);
        s_ai_cfg.alert_flag = FALSE;
        _voice_stat_update(IN_SILENCE);
    } else {
        // 一次触发模式，播放结束后，需要等待一段时间，再重新触发下一轮对话录音
        s_force_active = TRUE;
        PR_ERR("player stop, s_force_active = %d", s_force_active);
        tal_sw_timer_start(s_speech_active_timer, SPEECH_ACTIVE_TIMEOUT, TAL_TIMER_ONCE);
        PR_ERR("alert_flag = %d", s_ai_cfg.alert_flag);
        s_ai_cfg.alert_flag = FALSE;
        _voice_stat_update(IN_SILENCE);
    }

    return 0;
}
void ty_ai_toy_set_stat_start()
{
    _voice_stat_update(IN_START);
    return ;
}

void ty_ai_toy_set_stat_stop()
{
    _voice_stat_update(IN_STOP);
    return ;
}

void ty_ai_toy_set_trigger_mode_one_shot()
{
    s_ai_cfg.para.trigger_mode = TY_AI_TRIGGER_MODE_ONE_SHOT;
    return ;
}
// 语音活动状态超时后，强制进入SILENCE状态
INT_T ty_ai_toy_audio_frame_put(TKL_AUDIO_FRAME_INFO_T *pframe)
{
    // 按键检测
    //TUYA_GPIO_LEVEL_E read_level = TUYA_GPIO_LEVEL_HIGH;
    INT_T vad_active = 0;
    UINT32_T write_size = 0;
    //s_ai_cfg.para.audio_trigger_pin = TUYA_GPIO_NUM_47;
    //tkl_gpio_read(s_ai_cfg.para.audio_trigger_pin, &read_level);

    PR_ERR("trigger_mode = %d", s_ai_cfg.para.trigger_mode);
    if (s_ai_cfg.para.trigger_mode == TY_AI_TRIGGER_MODE_ONE_SHOT) {
        // FIXME: 目前固定是AUDIO_PCM_SLICE_TIME，否则需要按照10ms/20ms/30ms分片
        PR_ERR("process data = %d", pframe->buf_size);
        vad_active = vad_module->process(vad_module->self, pframe->pbuf, pframe->buf_size, NULL, 0);
        if (vad_active  < 0) {
            PR_ERR("vad error = %d", vad_active);
            return 1;
        }
        vad_active = s_vad_active; // 从回调函数中获取vad结果
    } else {
        // 持续模式，不需要vad检测
        PR_ERR (" vad_active =%d\n", vad_active);
        if (s_ai_cfg.voice_state == IN_STOP && !vad_active) {
            s_ai_cfg.frame_duration_ms = 0;
            return 1;
        }

        if (tuya_ai_mp3_is_playing() && vad_active) {
            PR_ERR("t5 mp3 is playing, stop it...");
            tuya_ai_mp3_player_stop();
            s_ai_cfg.wait_stop_play_threshold = WAIT_STOP_PLAY_THRESHOLD;
        }

        if (s_ai_cfg.voice_state == IN_STOP) {
            PR_ERR("move to voice silence...");
            _voice_stat_update(IN_SILENCE);
        }
    }

    //PR_ERR ("s_ai_cfg.voice_state = %d", s_ai_cfg.voice_state);
    switch (s_ai_cfg.voice_state) {
        case IN_SILENCE:
            if (vad_active) {
                if (s_ai_cfg.wait_stop_play_threshold >= AUDIO_PCM_SLICE_TIME) {
                    s_ai_cfg.wait_stop_play_threshold -= AUDIO_PCM_SLICE_TIME;
                    return 1;
                }

                ty_ai_audio_stream_write(s_ai_proc_handle, pframe->pbuf, pframe->buf_size);
                if (s_ai_cfg.frame_duration_ms == 0)
                PR_ERR("first voice frame...");
                // 需要等待Xms，才能切换到语音活动状态
                s_ai_cfg.frame_duration_ms += AUDIO_PCM_SLICE_TIME;
                // PR_ERR("frame duration = %d, active_threshold = %d", s_ai_cfg.frame_duration_ms, s_ai_cfg.active_threshold);
                if (s_ai_cfg.frame_duration_ms >= s_ai_cfg.active_threshold) {
                    // 之前是静音状态，现在是语音活动状态
                    _voice_stat_update(IN_START);
                    s_ai_cfg.frame_duration_ms = 0;
                }
            } else {
                // 静音状态
                s_ai_cfg.frame_duration_ms = 0;
                ty_ai_audio_stream_clear(s_ai_proc_handle);
            }
            break;
        case IN_START:
            PR_ERR("IN_START\n");
            _voice_stat_update(IN_VOICE);
        case IN_VOICE:
            PR_ERR("IN_VOICE\n");
            write_size = ty_ai_audio_stream_write(s_ai_proc_handle, pframe->pbuf, pframe->buf_size);
            // write_size == 0 means voice buffer full
            if (!vad_active || write_size == 0) {
                // 之前是语音活动状态，需要等待silence_threshold，才能切换到停止状态
                s_ai_cfg.frame_duration_ms += AUDIO_PCM_SLICE_TIME;
                if (s_ai_cfg.frame_duration_ms >= s_ai_cfg.silence_threshold || write_size == 0) {
                    if (write_size == 0)
                        PR_ERR("voice buffer full...");
                    // 之前是语音活动状态，切换到停止状态
                    // trigger timer immediately
                    tal_sw_timer_trigger(s_speech_active_timer);
                    _voice_stat_update(IN_STOP);
                    //PR_ERR("voice buffer size left = %d", ty_ai_audio_stream_get_size(s_ai_proc_handle));
                    s_ai_cfg.frame_duration_ms = 0;
                }
            } else {
                // 语音活动状态
                s_ai_cfg.frame_duration_ms = 0;
            }
            break;
        case IN_STOP:
            PR_ERR("IN_STOP\n");
            // 停止状态
            // do nothing
            break;
        default:
            break;
    }
    return 0;
}

STATIC INT_T _vad_callback(VOID *data, INT_T len, INT_T vad, VOID *userdata)
{
    INT_T *s_vad_active = (INT_T *)userdata;
    *s_vad_active = vad;
    return 0;
}

STATIC void _ty_ai_toy_task(void* arg)
{
    OPERATE_RET rt = 0;

    PR_ERR("ai toy task start");

    // create speech active timer
    TUYA_CALL_ERR_GOTO(tal_sw_timer_create(_speech_active_timer_timeout, NULL, &s_speech_active_timer), err);

    // subscribe t5 player event
    ty_subscribe_event(EVENT_T5_PLAYER, "ai_toy", _event_play_status_cb, SUBSCRIBE_TYPE_NORMAL);


    // create ns instance
    ns_config_t ns_cfg = {
        .sample_rate = AUDIO_SAMPLE_RATE,
        .level = NSLevel_Moderate,
    };
    TUYA_CALL_ERR_GOTO(ns_module->create(&ns_module->self, &ns_cfg), err);

    // create vad instance
    vad_config_t vad_para = {0};
    vad_para.sample_rate = AUDIO_SAMPLE_RATE;
    vad_para.mode = 3;
    TUYA_CALL_ERR_GOTO(vad_module->create(&vad_module->self, &vad_para), err);
    TUYA_CALL_ERR_GOTO(vad_module->set_callback(vad_module->self, _vad_callback, &s_vad_active), err);

    TUYA_CALL_ERR_GOTO(tuya_ai_mp3_player_init(), err);

    TY_AI_PROC_CONFIG_T ai_cfg = TY_AI_PROC_CFG_DEF;
    TUYA_CALL_ERR_GOTO(ty_ai_proc_create(&s_ai_proc_handle, &ai_cfg), err);

    PR_ERR("ty_ai_proc_create OK");

    tkl_thread_release(s_ty_ai_toy_hander);
    return;

err:
    if (ns_module) {
        ns_module->destroy(ns_module->self);
    }
    if (vad_module) {
        vad_module->destroy(vad_module->self);
    }
    PR_ERR("ai toy task exit");
    tkl_thread_release(s_ty_ai_toy_hander);
    return;
}

STATIC OPERATE_RET _report_sysinfo(VOID)
{
    // report volume dp to cloud
    PR_ERR("report volume dp to cloud");
    CHAR_T *devid = tuya_iot_get_gw_id();
    TY_OBJ_DP_S dp = {
        .dpid = 3,
        .type = PROP_VALUE,
        .value.dp_value = s_ai_cfg.volume,
    };
    return tuya_report_dp_async(devid, &dp, 1, NULL);
}

VOID ty_ai_toy_dp_cmd_cb(IN CONST TY_RECV_OBJ_DP_S *dp)
{
    for (UINT_T index = 0; index < dp->dps_cnt; index++) {
        // handle dp id = 3
        if (dp->dps[index].dpid == 3 && dp->dps[index].type == PROP_VALUE) {
            PR_ERR("SOC Rev DP Obj Cmd dpid:%d type:%d value:%d", dp->dps[index].dpid, dp->dps[index].type, dp->dps[index].value.dp_value);
            if (dp->dps[index].value.dp_value <= 100 && dp->dps[index].value.dp_value >= 0 && s_ai_cfg.volume != dp->dps[index].value.dp_value) {
                // update cfg
                s_ai_cfg.volume = dp->dps[index].value.dp_value;
                CHAR_T buf[32] = {0};
                snprintf(buf, sizeof(buf), "{\"volume\": %d}", s_ai_cfg.volume);
                wd_common_write(AI_TOY_PARA, (CONST BYTE_T *)buf, strlen(buf));
                tuya_ai_player_set_volume(s_ai_cfg.volume);
                _report_sysinfo();
            }
        }
    }
}

OPERATE_RET ty_ai_toy_init(void)
{
    PR_ERR("ai_toy_init");

    memset(&s_ai_cfg.para, 0x0, sizeof(TY_AI_TOY_PARA_T));
    if (s_ai_cfg.para.trigger_mode == TY_AI_TRIGGER_MODE_HOLD) {
        s_ai_cfg.silence_threshold = SILENCE_THRESHOLD_HOLD_MODE;
        s_ai_cfg.active_threshold = ACTIVE_THRESHOLD_HOLD_MODE;
    } else if (s_ai_cfg.para.trigger_mode == TY_AI_TRIGGER_MODE_ONE_SHOT) {
        s_ai_cfg.silence_threshold = SILENCE_THRESHOLD_ONE_SHORT_MODE;
        s_ai_cfg.active_threshold = ACTIVE_THRESHOLD_ONE_SHORT_MODE;
    } else {
        PR_ERR("trigger mode error");
        return OPRT_COM_ERROR;
    }
    s_ai_cfg.is_init = TRUE;
    return OPRT_OK;
}

OPERATE_RET ty_ai_toy_start(void)
{
    OPERATE_RET rt = OPRT_OK;
    PR_ERR("ai_toy_init");
    if (!s_ai_cfg.is_init) {
        PR_ERR("ai toy not init");
        return OPRT_COM_ERROR;
    }
    // create thread to handle ai toy
    rt = tkl_thread_create(&s_ty_ai_toy_hander, "ai_toy", 1024 * 400, THREAD_PRIO_2, _ty_ai_toy_task, NULL);
    if (rt != OPRT_OK) {
        PR_ERR("ty_ai_toy_start fail");
        return -1;
    }
    sleep (10);
    PR_ERR("ty_ai_toy_start success");

    return rt;
}
